/*
 * Copyright (c) 1999 Greg Haerr <greg@censoft.com>
 *
 * Test program for Micro-Windows
 * and DefWindowProc implementation required for window manager
 */
#include <stdio.h>
#include "windows.h"
#include "wintern.h"
#include "wintools.h"		/* Draw3dBox*/

#define APPCLASS	"test"

/* forward decls*/
LRESULT CALLBACK WndProc(HWND hwnd,UINT uMsg,WPARAM wp,LPARAM lp);

int
RegisterAppClass(void)
{
	WNDCLASS	wc;

	wc.style = CS_DBLCLKS | CS_VREDRAW | CS_HREDRAW;
	wc.lpfnWndProc = (WNDPROC)WndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = 0;
	wc.hIcon = 0; //LoadIcon(GetHInstance(), MAKEINTRESOURCE( 1));
	wc.hCursor = 0; //LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH)GetStockObject(GRAY_BRUSH);
	wc.lpszMenuName = NULL;
	wc.lpszClassName =  APPCLASS;
	return RegisterClass( &wc);
}

HWND
CreateAppWindow(void)
{
	HWND	hwnd;

	hwnd = CreateWindowEx(0L, APPCLASS,
		"Micro Application",
		WS_OVERLAPPEDWINDOW | WS_VISIBLE,
		CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		NULL, NULL, NULL, NULL);

	return hwnd;
}

LRESULT CALLBACK
WndProc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
	PAINTSTRUCT	ps;
	HDC		hdc;
	RECT		rc;
   
	switch( msg) {
	case WM_CREATE:
		break;

	case WM_DESTROY:
		//PostQuitMessage(0);
		break;

	case WM_SIZE:
		break;

	case WM_MOVE:
		break;

	case WM_SETFOCUS:
		PostMessage((HWND)wp, WM_PAINT, 0, 0L);
		break;

	case WM_KILLFOCUS:
		PostMessage((HWND)wp, WM_PAINT, 0, 0L);
		break;

	case WM_ERASEBKGND:
		if(GetFocus() != hwnd)
			return DefWindowProc(hwnd, msg, wp, lp);
		return 1;

	case WM_PAINT:
		hdc = BeginPaint(hwnd, &ps);

		if(GetFocus() == hwnd) {
			GetClientRect(hwnd, &rc);
			Draw3dShadow(hdc, rc.left, rc.top,
				rc.right-rc.left, rc.bottom-rc.top,
				GetSysColor(COLOR_3DDKSHADOW),
				GetSysColor(COLOR_3DLIGHT));
			InflateRect(&rc, -1, -1);
			FillRect(hdc, &rc, GetStockObject(GRAY_BRUSH));
		}

		SelectObject(hdc, GetStockObject(WHITE_PEN));
		MoveToEx(hdc, 2, 2, NULL);
		LineTo(hdc, 180, 230);
		SetBkColor(hdc, GRAY);
		TextOut(hdc, 2, 2, "Test", 4);
		EndPaint(hwnd, &ps);
		break;

	case WM_LBUTTONDOWN:
		break;

	case WM_MOUSEMOVE:
		break;

	case WM_LBUTTONUP:
		break;

	case WM_RBUTTONDOWN:
		WndRaiseWindow(hwnd);
		hdc = GetDC(hwnd);
		LineTo(hdc, 43, 150);
		ReleaseDC(hwnd, hdc);
		break;

	default:
		return DefWindowProc( hwnd, msg, wp, lp);
	}
	return( 0);
}

int WINAPI 
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine,
	int nShowCmd)
{
	MSG msg;

	RegisterAppClass();
	CreateAppWindow();
	CreateAppWindow();
	CreateAppWindow();
	CreateAppWindow();
	CreateAppWindow();
	CreateAppWindow();
	CreateAppWindow();
	CreateAppWindow();
	CreateAppWindow();

	while( GetMessage(&msg, NULL, 0, 0)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return 0;
}

LRESULT WINAPI
DefWindowProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	HDC		hdc;
	RECT		rc;
	DWORD		dwStyle;
	HBRUSH		hbr;
	PAINTSTRUCT	ps;
	POINT		curpt;
	int 		x, y;
	LPNCCALCSIZE_PARAMS lpnc;
	CHAR		szTitle[64];
	static POINT 	startpt;

	switch(msg) {
	case WM_NCCALCSIZE:
		/* calculate client rect from passed window rect in rgrc[0]*/
		lpnc = (LPNCCALCSIZE_PARAMS)lParam;
		dwStyle = GetWindowLong(hwnd, GWL_STYLE);
		if(dwStyle & WS_BORDER) {
			if((dwStyle & WS_CAPTION) == WS_CAPTION) {
				InflateRect(&lpnc->rgrc[0], -2, -2);
				lpnc->rgrc[0].top += SYSMETRICS_CYCAPTION;
			} else
				InflateRect(&lpnc->rgrc[0], -1, -1);
		}
		break;

	case WM_NCPAINT:
		/* repaint all non-client area*/
		dwStyle = GetWindowLong(hwnd, GWL_STYLE);
		if(dwStyle & WS_BORDER) {
			GetWindowRect(hwnd, &rc);
			hdc = GetWindowDC(hwnd);

			if((dwStyle & WS_CAPTION) == WS_CAPTION) {
				/* draw 2-line 3d border around window*/
				Draw3dBox(hdc, rc.left, rc.top,
					rc.right-rc.left, rc.bottom-rc.top,
					GetSysColor(COLOR_3DLIGHT),
					GetSysColor(COLOR_WINDOWFRAME));
				InflateRect(&rc, -1, -1);
				Draw3dBox(hdc, rc.left, rc.top,
					rc.right-rc.left, rc.bottom-rc.top,
					GetSysColor(COLOR_BACKGROUND),
					GetSysColor(COLOR_BACKGROUND));
				InflateRect(&rc, -1, -1);

				/* fill caption*/
				rc.bottom = rc.top + SYSMETRICS_CYCAPTION;
				hbr = CreateSolidBrush(GetFocus() == hwnd?
					GetSysColor(COLOR_ACTIVECAPTION):
					GetSysColor(COLOR_INACTIVECAPTION));
				FillRect(hdc, &rc, hbr);
				DeleteObject(hbr);

				/* draw caption text*/
				if(GetWindowText(hwnd, szTitle,
				   sizeof(szTitle))) {
					SetBkMode(hdc, TRANSPARENT);
					SetTextColor(hdc, WHITE);
					SelectObject(hdc,
					    GetStockObject(DEFAULT_GUI_FONT));
					GetWindowRect(hwnd, &rc);
					TextOut(hdc, rc.left+1, rc.top+1,
						szTitle, strlen(szTitle));
				}
			} else {
				SelectObject(hdc, GetStockObject(NULL_BRUSH));
				Rectangle(hdc, rc.left, rc.top, rc.right,
					rc.bottom);
			}

			ReleaseDC(hwnd, hdc);
		}
		break;

	case WM_NCHITTEST:
		dwStyle = GetWindowLong(hwnd, GWL_STYLE);
		if((dwStyle & WS_CAPTION) == WS_CAPTION) {
			POINTSTOPOINT(curpt, lParam);
			GetWindowRect(hwnd, &rc);
			InflateRect(&rc, -2, -2);
			rc.bottom = rc.top + SYSMETRICS_CYCAPTION;
			if(PtInRect(&rc, curpt))
				return HTCAPTION;
		}
		return HTCLIENT;

	case WM_NCLBUTTONDOWN:
		/* set focus on mouse down*/
		SetFocus(hwnd);

		/* start window drag if in caption area*/
		if(wParam == HTCAPTION) {
			POINTSTOPOINT(startpt, lParam);
			SetCapture(hwnd);
		}
		break;

	case WM_NCMOUSEMOVE:
		/* drag window with mousemove after mousedown*/
		if(GetCapture() == hwnd) {
			POINTSTOPOINT(curpt, lParam);
			x = curpt.x - startpt.x;
			y = curpt.y - startpt.y;
			GetWindowRect(hwnd, &rc);
			MoveWindow(hwnd, rc.left+x, rc.top+y, rc.right-rc.left,
				rc.bottom-rc.top, TRUE);
			startpt = curpt;
		}
		break;

	case WM_NCLBUTTONUP:
		/* stop window drag*/
		if(GetCapture() == hwnd)
			ReleaseCapture();
		break;

	case WM_NCRBUTTONDOWN:
		DestroyWindow(hwnd);
		break;

	case WM_GETTEXT:
		/* Get window text.  This routine requires
		 * knowledge of the internal window structure
		 */
		return strzcpy((LPSTR)lParam, hwnd->szTitle, wParam);

	case WM_ERASEBKGND:
		/* erase background with class background brush*/
		hbr = (HBRUSH)GetClassLong(hwnd, GCL_HBRBACKGROUND);
		if(!hbr)
			return 1;
		hdc = GetDC(hwnd);
		FillRect(hdc, NULL, hbr);
		ReleaseDC(hwnd, hdc);
		break;

	case WM_PAINT:
		/* required to send erasebkgnd for desktop window*/
		BeginPaint(hwnd, &ps);
		EndPaint(hwnd, &ps);
		break;
	}
	return 0;
}
