/* wingdi.h*/
/*
 * Copyright (c) 1999 Greg Haerr <greg@censoft.com>
 *
 * Win32 GDI structures and API
 */

/* portable coordinate definition*/
typedef COORD	GDICOORD;

typedef struct  tagRECT {
    GDICOORD left;
    GDICOORD top;
    GDICOORD right;
    GDICOORD bottom;
} RECT, NEAR *PRECT, FAR *LPRECT;

typedef struct  tagPOINT {
    GDICOORD x;
    GDICOORD y;
} POINT, NEAR *PPOINT, FAR *LPPOINT;

typedef struct  tagSIZE {
    GDICOORD cx;
    GDICOORD cy;
} SIZE, NEAR *PSIZE, FAR *LPSIZE;

HDC WINAPI	GetDC(HWND hWnd);
HDC WINAPI 	GetWindowDC(HWND hWnd);
int WINAPI 	ReleaseDC(HWND hWnd, HDC hDC);

typedef struct tagPAINTSTRUCT {
    HDC         hdc;
    BOOL        fErase;
    RECT        rcPaint;		/* nyi*/
    BOOL        fRestore;		/* nyi*/
    BOOL        fIncUpdate;		/* nyi*/
    BYTE        rgbReserved[32];
} PAINTSTRUCT, *PPAINTSTRUCT, *NPPAINTSTRUCT, FAR *LPPAINTSTRUCT;

HDC WINAPI 	BeginPaint(HWND hWnd, LPPAINTSTRUCT lpPaint);
BOOL WINAPI 	EndPaint(HWND hWnd, CONST PAINTSTRUCT *lpPaint);

/* Background Modes */
#define TRANSPARENT	1
#define OPAQUE		2

COLORREF WINAPI	SetTextColor(HDC, COLORREF);
COLORREF WINAPI	SetBkColor(HDC, COLORREF);
int WINAPI 	SetBkMode(HDC, int);

COLORREF WINAPI	SetPixel(HDC hdc, int x, int y, COLORREF crColor);
BOOL WINAPI 	MoveToEx(HDC hdc, int x, int y, LPPOINT lpPoint);
BOOL WINAPI 	LineTo(HDC, int, int);
BOOL WINAPI	Rectangle(HDC hdc, int nLeft, int nTop, int nRight,int nBottom);
int WINAPI	FillRect(HDC hDC, CONST RECT *lprc, HBRUSH hbr);

/* ExTextOut options*/
#define ETO_OPAQUE	0x0002
#define ETO_CLIPPED	0x0004		/* nyi*/

BOOL WINAPI	TextOut(HDC hdc, int x, int y, LPCSTR lpszString, int cbString);
BOOL WINAPI	ExtTextOut(HDC hdc, int x, int y, UINT fuOptions,
			CONST RECT *lprc, LPCSTR lpszString, UINT cbCount,
			CONST INT *lpDx);

/* GetSysColor, FillRect colors*/
#define COLOR_SCROLLBAR         0
#define COLOR_BACKGROUND        1
#define COLOR_ACTIVECAPTION     2
#define COLOR_INACTIVECAPTION   3
#define COLOR_MENU              4
#define COLOR_WINDOW            5
#define COLOR_WINDOWFRAME       6
#define COLOR_MENUTEXT          7
#define COLOR_WINDOWTEXT        8
#define COLOR_CAPTIONTEXT       9
#define COLOR_ACTIVEBORDER      10
#define COLOR_INACTIVEBORDER    11
#define COLOR_APPWORKSPACE      12
#define COLOR_HIGHLIGHT         13
#define COLOR_HIGHLIGHTTEXT     14
#define COLOR_BTNFACE           15
#define COLOR_BTNSHADOW         16
#define COLOR_GRAYTEXT          17
#define COLOR_BTNTEXT           18
#define COLOR_INACTIVECAPTIONTEXT 19
#define COLOR_BTNHIGHLIGHT      20
#define COLOR_3DDKSHADOW        21
#define COLOR_3DLIGHT           22
#define COLOR_INFOTEXT          23
#define COLOR_INFOBK            24
#define COLOR_DESKTOP           COLOR_BACKGROUND
#define COLOR_3DFACE            COLOR_BTNFACE
#define COLOR_3DSHADOW          COLOR_BTNSHADOW
#define COLOR_3DHIGHLIGHT       COLOR_BTNHIGHLIGHT
#define COLOR_3DHILIGHT         COLOR_BTNHIGHLIGHT
#define COLOR_BTNHILIGHT        COLOR_BTNHIGHLIGHT
#define COLOR_ALTERNATEBUTTONFACE	25
#define COLOR_HOTLIGHT                  26
#define COLOR_GRADIENTACTIVECAPTION     27
#define COLOR_GRADIENTINACTIVECAPTION   28

DWORD WINAPI	GetSysColor(int nIndex);

/* Stock Logical Objects */
#define WHITE_BRUSH         0
#define LTGRAY_BRUSH        1
#define GRAY_BRUSH          2
#define DKGRAY_BRUSH        3
#define BLACK_BRUSH         4
#define NULL_BRUSH          5
#define HOLLOW_BRUSH        NULL_BRUSH
#define WHITE_PEN           6
#define BLACK_PEN           7
#define NULL_PEN            8
#define OEM_FIXED_FONT      10
#define ANSI_FIXED_FONT     11
#define ANSI_VAR_FONT       12
#define SYSTEM_FONT         13
#define DEVICE_DEFAULT_FONT 14
#define DEFAULT_PALETTE     15
#define SYSTEM_FIXED_FONT   16
#define DEFAULT_GUI_FONT    17
#define DC_BRUSH            18
#define DC_PEN              19
#define STOCK_LAST          19

/* Object types*/
#define OBJ_PEN             1
#define OBJ_BRUSH           2
#define OBJ_DC              3
#define OBJ_METADC          4
#define OBJ_PAL             5
#define OBJ_FONT            6
#define OBJ_BITMAP          7
#define OBJ_REGION          8
#define OBJ_METAFILE        9
#define OBJ_MEMDC           10
#define OBJ_EXTPEN          11
#define OBJ_ENHMETADC       12
#define OBJ_ENHMETAFILE     13

HGDIOBJ WINAPI	GetStockObject(int nObject);
HGDIOBJ WINAPI	SelectObject(HDC hdc, HGDIOBJ hObject);
BOOL WINAPI	DeleteObject(HGDIOBJ hObject);

/* Brush Styles */
#define BS_SOLID            0
#define BS_NULL             1
#define BS_HOLLOW           BS_NULL

HBRUSH WINAPI	CreateSolidBrush(COLORREF crColor);

/* Pen Styles */
#define PS_SOLID            0
#define PS_NULL             5

HPEN WINAPI	CreatePen(int nPenStyle, int nWidth, COLORREF crColor);
