/*
 * Copyright (c) 1999 Greg Haerr <greg@censoft.com>
 *
 * Main module of micro-Windows
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#if UNIX
#include <unistd.h>
#include <sys/time.h>
#endif
#if ELKS
#include <linuxmt/posix_types.h>
#include <linuxmt/time.h>
#endif
#include "windows.h"
#include "wintern.h"

/*
 * External definitions defined here.
 */
HWND		listwp;			/* list of all windows */
HWND		rootwp;			/* root window pointer */
HWND		focuswp;		/* focus window for keyboard */
HWND		mousewp;		/* window mouse is currently in */
HWND		capturewp;		/* capture window*/
HCURSOR		curcursor;		/* currently enabled cursor */
COORD		cursorx;		/* current x position of cursor */
COORD		cursory;		/* current y position of cursor */
SCREENINFO	sinfo;			/* screen information */
int		keyb_fd;		/* the keyboard file descriptor */
int		mouse_fd;		/* the mouse file descriptor */

int
main(int ac,char **av)
{
	if(GsOpen() < 0)
		exit(1);

	/* call windows main program entry point*/
	WinMain(NULL, NULL, NULL, SW_SHOW);

	GsClose();
	return 0;
}

/*
 * Open a connection from a new client to the server.
 * Returns -1 on failure.
 */
int
GsOpen(void)
{
	/* Client calls this routine once.  We 
	 * init everything here
	 */
	if(GsInitialize() < 0)
		return -1;
        return 1;
}

/*
 * Close the connection to the server.
 */
void
GsClose(void)
{
	GsTerminate();
}

#if MSDOS
void
GsSelect(void)
{
	/* If mouse data present, service it*/
	if(mousedev.Poll())
		GsCheckMouseEvent();

	/* If keyboard data present, service it*/
	if(kbddev.Poll())
		GsCheckKeyboardEvent();

}
#endif

#if UNIX
void
GsSelect(void)
{
	fd_set	rfds;
	int 	e;
	struct timeval to;

	/* Set up the FDs for use in the main select(): */
	FD_ZERO(&rfds);
	FD_SET(mouse_fd, &rfds);
	FD_SET(keyb_fd, &rfds);

	/* Set up the timeout for the main select().  If
	 * the mouse is captured we're probably moving a window,
	 * so poll quickly to allow other windows to repaint while
	 * checking for more event input.
	 */
	to.tv_sec = 0L;
	if(capturewp)
		to.tv_usec = 1L;
	else to.tv_usec = 10000L;

	/* Wait for some input on any of the fds in the set or a timeout: */
	if((e = select(16, &rfds, NULL, NULL, &to)) > 0) {
		
		/* If data is present on the mouse fd, service it: */
		if(FD_ISSET(mouse_fd, &rfds))
			GsCheckMouseEvent();

		/* If data is present on the keyboard fd, service it: */
		if(FD_ISSET(keyb_fd, &rfds))
			GsCheckKeyboardEvent();

	} 
	else if(!e) {
#if BOGL
		if(bogl_CheckVtChange())
			WndRedrawScreen();
#endif
	} else
		if(errno != EINTR)
			perror("Select() call in main failed");
}
#endif

/*
 * Initialize the graphics and mouse devices at startup.
 * Returns nonzero with a message printed if the initialization failed.
 */
int
GsInitialize(void)
{
	HWND		wp;		/* root window */
	WNDCLASS	wc;
	static SWCURSOR arrow = {	/* default arrow cursor*/
		16, 16,
		0,  0,
		WHITE, BLACK,
		{ 0xe000, 0x9800, 0x8600, 0x4180,
		  0x4060, 0x2018, 0x2004, 0x107c,
		  0x1020, 0x0910, 0x0988, 0x0544,
		  0x0522, 0x0211, 0x000a, 0x0004 },
		{ 0xe000, 0xf800, 0xfe00, 0x7f80,
		  0x7fe0, 0x3ff8, 0x3ffc, 0x1ffc,
		  0x1fe0, 0x0ff0, 0x0ff8, 0x077c,
		  0x073e, 0x021f, 0x000e, 0x0004 }
	};
	extern LISTHEAD classHead;

	wp = ItemNew(struct hwnd);
	if (!wp) {
		perror("No memory for root window");
		return -1;
	}

	if ((keyb_fd = GdOpenKeyboard()) < 0) {
		perror("Cannot initialise keyboard");
		ItemFree(wp);
		return -1;
	}

	if (GdOpenScreen() < 0) {
		perror("Cannot initialise screen");
		GdCloseKeyboard();
		ItemFree(wp);
		return -1;
	}

	if ((mouse_fd = GdOpenMouse()) < 0) {
		perror("Cannot initialise mouse");
		GdCloseScreen();
		GdCloseKeyboard();
		ItemFree(wp);
		return -1;
	}

	/*
	 * Get screen dimensions for our own use.
	 */
	GdGetScreenInfo(&sinfo);
	GdGetModifierInfo(&sinfo.modifiers);
	GdGetButtonInfo(&sinfo.buttons);

	/*
	 * Initialize the root window.
	 */
	wc.style = CS_DBLCLKS | CS_VREDRAW | CS_HREDRAW;
	wc.lpfnWndProc = (WNDPROC)DefWindowProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = 0;
	wc.hIcon = 0; //LoadIcon(GetHInstance(), MAKEINTRESOURCE( 1));
	wc.hCursor = 0; //LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = CreateSolidBrush(RGB(0,128,128));
	wc.lpszMenuName = NULL;
	wc.lpszClassName =  "DeskTop";
	RegisterClass( &wc);
	
	wp->style = WS_CAPTION | WS_VISIBLE;
	wp->exstyle = 0;
	wp->pClass = (PWNDCLASS)classHead.head;
	wp->parent = NULL;
	wp->children = NULL;
	wp->siblings = NULL;
	wp->next = NULL;
	SetRect(&wp->winrect, 0, 0, sinfo.cols, sinfo.rows);
	WndCalcClientRect(wp, wp->style);
	wp->cursor = NULL;
	wp->unmapcount = 0;
	wp->id = 0;
	strcpy(wp->szTitle, "Micro-Windows");
	wp->gotPaintMsg = 0;

	listwp = wp;
	rootwp = wp;
	focuswp = wp;
	mousewp = wp;

	/* schedule desktop window paint*/
	PostMessage(rootwp, WM_PAINT, 0, 0L);

#if BOGL
	bogl_InitVt();
#endif

	/*
	 * Initialize and position the default cursor.
	 */
	curcursor = NULL;
	cursorx = -1;
	cursory = -1;
	GsMoveCursor(sinfo.cols / 2, sinfo.rows / 2);
	WndSetCursor(rootwp, &arrow);

	/*
	 * Finally tell the mouse driver some things.
	 */
	/*GdSetAccelMouse(5, 3);*/
	GdRestrictMouse(0, 0, sinfo.cols - 1, sinfo.rows - 1);
	GdMoveMouse(sinfo.cols / 2, sinfo.rows / 2);

	return 0;
}

/*
 * Here to close down the server.
 */
void
GsTerminate(void)
{
	GdCloseScreen();
	GdCloseMouse();
	GdCloseKeyboard();
#if BOGL
	bogl_RedrawVt(vterm);
#endif
	exit(0);
}
