/*
 * Copyright (c) 1999 Greg Haerr <greg@censoft.com>
 *
 * Win32 API upper level window creation, management and msg routines
 */
#include "windows.h"
#include "wintern.h"
#include <stdlib.h>
#include <string.h>

#define PAINTONCE	1	/* =1 to queue paint msgs only once*/

LISTHEAD msgHead;		/* application msg queue*/
LISTHEAD classHead;		/* register class list*/

int	SYSMETRICS_CYCAPTION = 12;

LRESULT WINAPI
SendMessage(HWND hwnd, UINT Msg,WPARAM wParam,LPARAM lParam)
{
	if(hwnd && hwnd->pClass)
		return (*hwnd->pClass->lpfnWndProc)(hwnd, Msg, wParam, lParam);
	return 0;
}

BOOL WINAPI
PostMessage(HWND hwnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	MSG *	pMsg;

#if PAINTONCE
	/* don't queue paint msgs, just set window bit*/
	if(Msg == WM_PAINT) {
		hwnd->gotPaintMsg = 1;
		return TRUE;
	}
#endif
#if MOUSETEST
	if(Msg == WM_MOUSEMOVE) {
		PLIST	p;
		for(p=msgHead.head; p; p=p->next) {
			pMsg = ItemAddr(p, MSG, link);
			if(pMsg->hwnd == hwnd && pMsg->message == Msg) {
				pMsg->wParam = wParam;
				pMsg->lParam = lParam;
				pMsg->time = 0;
				pMsg->pt.x = cursorx;
				pMsg->pt.y = cursory;
				return TRUE;
			}
		}
	}
#endif
	pMsg = ItemNew(MSG);
	if(!pMsg)
		return FALSE;
	pMsg->hwnd = hwnd;
	pMsg->message = Msg;
	pMsg->wParam = wParam;
	pMsg->lParam = lParam;
	pMsg->time = 0;
	pMsg->pt.x = cursorx;
	pMsg->pt.y = cursory;
	ListAdd(&msgHead, &pMsg->link);
	return TRUE;
}

VOID WINAPI
PostQuitMessage(int nExitCode)
{
	PostMessage(NULL, WM_QUIT, nExitCode, 0L);
}

BOOL WINAPI
GetMessage(LPMSG lpMsg,HWND hwnd,UINT wMsgFilterMin,UINT wMsgFilterMax)
{
	PMSG	pNxtMsg;
	HWND	wp;

	/* check if no messages in queue*/
	while(msgHead.head == NULL) {
#if PAINTONCE
		/* check all windows for pending paint messages*/
		for(wp=listwp; wp; wp=wp->next) {
			/*
			 * Tricky: only repaint window if there
			 * isn't a mouse capture (window move) in progress,
			 * or, the window is the moving window.
			 */
			if(wp->gotPaintMsg 
			   && (!capturewp || capturewp == wp)) {
		paint:
				wp->gotPaintMsg = 0;
				lpMsg->hwnd = wp;
				lpMsg->message = WM_PAINT;
				lpMsg->wParam = 0;
				lpMsg->lParam = 0;
				lpMsg->time = 0;
				lpMsg->pt.x = cursorx;
				lpMsg->pt.y = cursory;
				return TRUE;
			} else if(wp->gotPaintMsg && capturewp) {
				/* All other windows we'll check for
				 * event input first, then allow repaint.
				 */
				GsSelect();
				if(msgHead.head == NULL)
					goto paint;
			}
		}
#endif
		GsSelect();
	}

	pNxtMsg = (PMSG)msgHead.head;
	ListRemove(&msgHead, &pNxtMsg->link);
	*lpMsg = *pNxtMsg;
	ItemFree(pNxtMsg);
	return lpMsg->message != WM_QUIT;
}

BOOL WINAPI 
TranslateMessage(CONST MSG *lpMsg)
{
	return FALSE;
}

LONG WINAPI
DispatchMessage(CONST MSG *lpMsg)
{
	return SendMessage(lpMsg->hwnd, lpMsg->message, lpMsg->wParam,
		lpMsg->lParam);
}

/* find the registered window class struct by name*/
PWNDCLASS
WndFindClassByName(LPCSTR lpClassName)
{
	PLIST		p;
	PWNDCLASS	pClass;

	for(p=classHead.head; p; p=p->next) {
		pClass = ItemAddr(p, WNDCLASS, link);
		if(strcmpi(pClass->szClassName, lpClassName) == 0)
			return pClass;
	}
	return NULL;
}

ATOM WINAPI
RegisterClass(CONST WNDCLASS *lpWndClass)
{
	PWNDCLASS	pClass;

	/* check if already present*/
	pClass = WndFindClassByName(lpWndClass->lpszClassName);
	if(pClass)
		return 0;
	
	/* copy class into new struct*/
	pClass = ItemNew(WNDCLASS);
	if(!pClass)
		return 0;
	*pClass = *lpWndClass;
	strcpy(pClass->szClassName, lpWndClass->lpszClassName);
	ListAdd(&classHead, &pClass->link);

	return 1;
}

BOOL WINAPI
UnregisterClass(LPCSTR lpClassName, HINSTANCE hInstance)
{
	PWNDCLASS	pClass;

	pClass = WndFindClassByName(lpClassName);
	if(!pClass)
		return FALSE;
	ListRemove(&classHead, &pClass->link);
	DeleteObject(pClass->hbrBackground);
	ItemFree(pClass);
	return TRUE;
}

HWND WINAPI
CreateWindowEx(DWORD dwExStyle, LPCSTR lpClassName, LPCSTR lpWindowName,
	DWORD dwStyle, int x, int y, int nWidth, int nHeight,
	HWND hwndParent, HMENU hMenu, HINSTANCE hInstance, LPVOID lpParam)
{
	HWND		pwp;		/* parent window */
	HWND		wp;		/* new window */
	PWNDCLASS	pClass;
	LPCREATESTRUCT	lpcs;
	static int	nextx = 20;
	static int	nexty = 20;

	pClass = WndFindClassByName(lpClassName);
	if(!pClass)
		return NULL;

	if(x == CW_USEDEFAULT || y == CW_USEDEFAULT) {
		x = nextx;
		nextx += 10;
		y = nexty;
		nexty += 10;
		if(nextx > 200)
			nextx = nexty = 20;
	}
	if(nWidth == CW_USEDEFAULT || nHeight == CW_USEDEFAULT) {
		nWidth = 250;
		nHeight = 250;
	}

	if(hwndParent == NULL)
		hwndParent = rootwp;
	pwp = hwndParent;

	wp = ItemNew(struct hwnd);
	if(!wp)
		return NULL;

	wp->pClass = pClass;
	wp->style = dwStyle;
	wp->exstyle = dwExStyle;
	wp->parent = pwp;
	wp->children = NULL;
	wp->siblings = pwp->children;
	wp->next = listwp;
	wp->winrect.left = x;	// pwp->x + x;
	wp->winrect.top = y;	// pwp->y + y;
	wp->winrect.right = x + nWidth;
	wp->winrect.bottom = y + nHeight;
	wp->cursor = pwp->cursor;
	wp->cursor->usecount++;
	wp->unmapcount = pwp->unmapcount + 1;
	wp->id = (int)hMenu;
	wp->gotPaintMsg = 0;
	strzcpy(wp->szTitle, lpWindowName, sizeof(wp->szTitle));
	pwp->children = wp;
	listwp = wp;

	/* calculate client area*/
	WndCalcClientRect(wp, dwStyle);

	/* note: the following assignment may have to change
	 * if the WINAPI decl changes from cdecl
	 */
	lpcs = (LPCREATESTRUCT)&dwExStyle;
	if(SendMessage(wp, WM_CREATE, 0, (LPARAM)lpcs) == -1) {
		WndDestroyWindow(wp, FALSE);
		return NULL;
	}

	/* send SIZE and MOVE msgs*/
	WndSendSizeMove(wp, TRUE, TRUE);

	if(wp->style & WS_VISIBLE) {
		WndShowWindow(wp);
		SetFocus(wp);
	}

	return wp;
}

BOOL WINAPI
DestroyWindow(HWND hwnd)
{
	WndDestroyWindow(hwnd, TRUE);
	return TRUE;
}

/*
 * Destroy the specified window, and all of its children.
 * This is a recursive routine.
 */
void
WndDestroyWindow(HWND hwnd,BOOL bSendMsg)
{
	HWND	wp = hwnd;
	HWND	prevwp;
	PLIST	p;
	PMSG	pmsg;

	if (wp == rootwp)
		return;

	/*
	 * Unmap the window.
	 */
	if (wp->unmapcount == 0)
		WndHideWindow(wp, TRUE);

	if(bSendMsg)
		SendMessage(hwnd, WM_DESTROY, 0, 0L);

	/*
	 * Disable all sendmessages to this window.
	 */
	wp->pClass = NULL;

	/*
	 * Destroy all children, sending WM_DESTROY messages.
	 */
	while (wp->children)
		WndDestroyWindow(wp->children, bSendMsg);

	/*
	 * Free any cursor associated with the window.
	 */
	if (wp->cursor->usecount-- == 1) {
		free(wp->cursor);
		wp->cursor = NULL;
	}

	/*
	 * Remove this window from the child list of its parent.
	 */
	prevwp = wp->parent->children;
	if (prevwp == wp)
		wp->parent->children = wp->siblings;
	else {
		while (prevwp->siblings != wp)
			prevwp = prevwp->siblings;
		prevwp->siblings = wp->siblings;
	}
	wp->siblings = NULL;

	/*
	 * Remove this window from the complete list of windows.
	 */
	prevwp = listwp;
	if (prevwp == wp)
		listwp = wp->next;
	else {
		while (prevwp->next != wp)
			prevwp = prevwp->next;
		prevwp->next = wp->next;
	}
	wp->next = NULL;

	/*
	 * Forget various information related to this window.
	 * Then finally free the structure.
	 */

	/* Remove all messages from msg queue for this window*/
	for(p=msgHead.head; p; ) {
		pmsg = ItemAddr(p, MSG, link);
		if(pmsg->hwnd == wp) {
			p = p->next;
			ListRemove(&msgHead, &pmsg->link);
		} else
			p = p->next;
	}

	// FIXME: destroy hdc's relating to window?

	if (wp == capturewp) {
		capturewp = NULL;
		GsCheckMouseWindow();
	}

	if (wp == focuswp)
		SetFocus(rootwp);

	ItemFree(wp);
}

BOOL WINAPI
ShowWindow(HWND hwnd, int nCmdShow)
{
	if(!hwnd)
		return FALSE;
	
	// fix: send show msg

	switch(nCmdShow) {
	case SW_HIDE:
		if (!(hwnd->style & WS_VISIBLE))
			return FALSE;
		WndHideWindow(hwnd, TRUE);
		hwnd->style &= ~WS_VISIBLE;
		break;

	default:
		if (hwnd->style & WS_VISIBLE)
			return FALSE;
		hwnd->style |= WS_VISIBLE;
		WndShowWindow(hwnd);
	}
	return TRUE;
}

BOOL WINAPI
UpdateWindow(HWND hwnd)
{
#if PAINTONCE
	if(hwnd && hwnd->gotPaintMsg) {
		SendMessage(hwnd, WM_PAINT, 0, 0L);
		hwnd->gotPaintMsg = 0;
		return TRUE;
	}
	return FALSE;
#else
	// fix: remove other paint messages from queue
	SendMessage(hwnd, WM_PAINT, 0, 0L);
	return TRUE;
#endif
}

HWND WINAPI
GetFocus(VOID)
{
	return focuswp;
}

HWND WINAPI
SetFocus(HWND hwnd)
{
	HWND	oldfocus;

	/* if NULL or hidden, set focus to desktop*/
	if(!hwnd || hwnd->unmapcount)
		hwnd = rootwp;

	if (hwnd == focuswp)
		return focuswp;

	// fix: ncpaint must go up child chain
	SendMessage(focuswp, WM_KILLFOCUS, (WPARAM)hwnd, 0L);

	oldfocus = focuswp;
	focuswp = hwnd;
	// fix: ncpaint must go up child chain
	SendMessage(focuswp, WM_SETFOCUS, (WPARAM)oldfocus, 0L);

	SendMessage(oldfocus, WM_NCPAINT, 0, 0L);
	SendMessage(focuswp, WM_NCPAINT, 0, 0L);

	return oldfocus;
}

/* set the client rect for a window from the window position and style*/
void
WndCalcClientRect(HWND hwnd, DWORD dwStyle)
{
	NCCALCSIZE_PARAMS	nccs;

	/* set first rectangle to window rect*/
	nccs.rgrc[0] = hwnd->winrect;
	SendMessage(hwnd, WM_NCCALCSIZE, FALSE, (LPARAM)&nccs);
	hwnd->clirect = nccs.rgrc[0];
}

BOOL WINAPI
GetClientRect(HWND hwnd, LPRECT lpRect)
{
	if(!hwnd || !lpRect)
		return FALSE;

	/* convert client area rect from screen coordinates*/
	lpRect->left = 0;
	lpRect->top = 0;
	lpRect->right = hwnd->clirect.right - hwnd->clirect.left;
	lpRect->bottom = hwnd->clirect.bottom - hwnd->clirect.top;
	return TRUE;
}

BOOL WINAPI
GetWindowRect(HWND hwnd, LPRECT lpRect)
{
	if(!hwnd || !lpRect)
		return FALSE;
	
	/* window rect is already in screen coordinates*/
	*lpRect = hwnd->winrect;
	return TRUE;
}

BOOL WINAPI
ClientToScreen(HWND hwnd, LPPOINT lpPoint)
{
	if(!hwnd || !lpPoint)
		return FALSE;
	lpPoint->x += hwnd->clirect.left;
	lpPoint->y += hwnd->clirect.top;
	return TRUE;
}

BOOL WINAPI
ScreenToClient(HWND hwnd, LPPOINT lpPoint)
{
	if(!hwnd || !lpPoint)
		return FALSE;
	lpPoint->x -= hwnd->clirect.left;
	lpPoint->y -= hwnd->clirect.top;
	return TRUE;
}

BOOL WINAPI
SetRect(LPRECT lprc, int xLeft, int yTop, int xRight, int yBottom)
{
	lprc->left = xLeft;
	lprc->top = yTop;
	lprc->right = xRight;
	lprc->bottom = yBottom;
	return TRUE;
}

BOOL WINAPI
InflateRect(LPRECT lprc, int dx, int dy)
{
	lprc->left -= dx;
	lprc->top -= dy;
	lprc->right += dx;
	lprc->bottom += dy;
	return TRUE;
}

BOOL WINAPI
PtInRect(CONST RECT *lprc, POINT pt)
{
	return (pt.x >= lprc->left && pt.x < lprc->right &&
		pt.y >= lprc->top && pt.y < lprc->bottom);
}

LONG WINAPI
GetWindowLong(HWND hwnd, int nIndex)
{
	switch(nIndex) {
	case GWL_STYLE:
		return hwnd->style;
	case GWL_EXSTYLE:
		return hwnd->exstyle;
	}
	return 0;
}

DWORD WINAPI
GetClassLong(HWND hwnd, int nIndex)
{
	switch(nIndex) {
	case GCL_HBRBACKGROUND:
		return (DWORD)hwnd->pClass->hbrBackground;
	}
	return 0;
}

int WINAPI
GetWindowText(HWND hwnd, LPSTR lpString, int nMaxCount)
{
	return SendMessage(hwnd, WM_GETTEXT, nMaxCount, (LPARAM)lpString);
}

BOOL WINAPI
MoveWindow(HWND hwnd, int x, int y, int nWidth, int nHeight, BOOL bRepaint)
{
	int		hidden;
	BOOL		bMove, bSize;

	if(!hwnd || hwnd == rootwp)
		return FALSE;

	// disallow width, height <= 0?

	//x += hwnd->parent->x;
	//y += hwnd->parent->y;

	bMove = (hwnd->winrect.left != x || hwnd->winrect.top != y);
	bSize = ((hwnd->winrect.right - hwnd->winrect.left) != nWidth ||
		 (hwnd->winrect.bottom - hwnd->winrect.top) != nHeight);
	if(!bMove && !bSize)
		return TRUE;

	/* could optimize to not require redraw when possible*/
	hidden = hwnd->unmapcount;
	if(!hidden)
		WndHideWindow(hwnd, FALSE);
	hwnd->winrect.left = x;
	hwnd->winrect.top = y;
	hwnd->winrect.right = x + nWidth;
	hwnd->winrect.bottom = y + nHeight;
	WndCalcClientRect(hwnd, hwnd->style);
	WndSendSizeMove(hwnd, bSize, bMove);

	if(!hidden)
		WndShowWindow(hwnd);

	return TRUE;
}

void
WndSendSizeMove(HWND hwnd, BOOL bSize, BOOL bMove)
{
	DWORD	dwStyle;
	RECT	rc;

	if(bSize) {
		GetClientRect(hwnd, &rc);
		SendMessage(hwnd, WM_SIZE, SIZE_RESTORED,
			MAKELONG(rc.right, rc.bottom));
	}
	if(bMove) {
		dwStyle = GetWindowLong(hwnd, GWL_STYLE);
		if(dwStyle & WS_CHILD)
			GetClientRect(hwnd->parent, &rc);
		else
			GetWindowRect(hwnd, &rc);
		SendMessage(hwnd, WM_MOVE, 0, MAKELONG(rc.left, rc.top));
	}
}

/*
 * Specify a cursor for a window.
 * This cursor will only be used within that window, and by default
 * for its new children.  If the cursor is currently within this
 * window, it will be changed to the new one immediately.
 */
void
WndSetCursor(HWND wp, PSWCURSOR pcursor)
{
	HCURSOR	cp;
	int	bytes;

	if(!wp || !pcursor)
		return;

	bytes = IMAGE_SIZE(pcursor->width,pcursor->height) * sizeof(IMAGEBITS);

	/*
	 * See if the window is using a shared cursor definition.
	 * If so, then allocate a new private one, otherwise reuse it.
	 */
	cp = wp->cursor;
	if (!cp || cp->usecount-- > 1) {
		cp = ItemNew(struct hcursor);
		if(!cp)
			return;
	}

	cp->usecount = 1;
	cp->cursor.width = pcursor->width;
	cp->cursor.height = pcursor->height;
	cp->cursor.hotx = pcursor->hotx;
	cp->cursor.hoty = pcursor->hoty;
	cp->cursor.fgcolor = pcursor->fgcolor;
	cp->cursor.bgcolor = pcursor->bgcolor;
	memcpy(cp->cursor.image, pcursor->image, bytes);
	memcpy(cp->cursor.mask, pcursor->mask, bytes);
	wp->cursor = cp;

	/*
	 * If this was the current cursor, then draw the new one.
	 */
	if (cp == curcursor || curcursor == NULL) {
		GdMoveCursor(cursorx - cp->cursor.hotx,
			cursory - cp->cursor.hoty);
		GdSetCursor(&cp->cursor);
	}
}

HWND WINAPI
GetCapture(VOID)
{
	return capturewp;
}

HWND WINAPI
SetCapture(HWND hwnd)
{
	HWND	oldCapture = capturewp;

	capturewp = hwnd;
	GsCheckMouseWindow();
	return oldCapture;
}

BOOL WINAPI
ReleaseCapture(VOID)
{
	capturewp = NULL;
	GsCheckMouseWindow();
	return TRUE;
}
